﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using MFORMATSLib;

using Emgu.CV;
using Emgu.CV.Structure;

namespace SampleMFormatsOpenCV
{
    public partial class Form1 : Form
    {
        private MFPreviewClass m_objPreview;
        private MFLiveClass m_objLive;
        private Thread m_threadWorker;	//Working thread
        private bool m_bWork;

        public Form1()
        {
            InitializeComponent();
        }

        private CancellationTokenSource cancelSource;
        private void Form1_Load(object sender, EventArgs e)
        {
            m_objPreview = new MFPreviewClass();
            try
            {
                m_objLive = new MFLiveClass();
            }
            catch (Exception ex)
            {
                MessageBox.Show("MLive device not found\n\n" + ex.ToString());
            }
            //Configure preview
            m_objPreview.PreviewWindowSet("", panelPreview.Handle.ToInt32());
            m_objPreview.PreviewEnable("", 1, 1);

            //Fill live devices
            FillMLiveDevices();

            // Create & config background worker
            cancelSource = new CancellationTokenSource();
            m_threadWorker = new Thread(() => thread_DoWork(cancelSource.Token));
            m_threadWorker.Name = "thread_DoWork";
			m_threadWorker.Start();

            System.Windows.Forms.Control.CheckForIllegalCrossThreadCalls = false;   //Need for use renderersCombobox in multithreading 
        }


        private void thread_DoWork(CancellationToken token)
        {
            MFFactoryClass pFactory = new MFFactoryClass();

            // Don't forget to copy a corresponding XML file for a cascade classifier
            CascadeClassifier cascadeClassifier = new CascadeClassifier(@"haarcascade_frontalface_alt2.xml");

            while (!token.IsCancellationRequested)
            {
                if (m_bWork)
                {
                    MFFrame pFrame = null;

                    try
                    {
                        m_objLive.SourceFrameGet(-1, out pFrame, "");

                        MFFrame pFrameDraw = null;
                        pFrame.MFClone(out pFrameDraw, eMFrameClone.eMFC_Video_ForceCPU, eMFCC.eMFCC_ARGB32);
                        Image<Rgba, Byte> cvFrame = GetCVImageFromMFFrame(pFrameDraw);

                        // Recognize faces and detect their bounding boxes
                        int nSizeFactor = Math.Min(cvFrame.Height, cvFrame.Width);
                        Size szMinSize = new Size((int)(nSizeFactor * 0.2), (int)(nSizeFactor * 0.2));
                        Size szMaxSize = new Size((int)(nSizeFactor * 0.35), (int)(nSizeFactor * 0.35));
                        Rectangle[] rcFacesDetected = cascadeClassifier.DetectMultiScale(cvFrame, 1.1, 5, szMinSize, szMaxSize);

                        // Draw corresponding bounding boxes
                        foreach( Rectangle rcFace in rcFacesDetected)
                            cvFrame.Draw(rcFace, new Rgba(255, 0, 0, 255), 3);

                        // Send frame to the preview
                        m_objPreview.ReceiverFramePut(pFrameDraw, -1, "");

                        // Release frame - DO NOT FORGOT TO DO THIS !!!
                        Marshal.ReleaseComObject(pFrame);
                        Marshal.ReleaseComObject(pFrameDraw);
                    }
                    catch
                    {
                    }
                }
                else
                {
                    Thread.Sleep(1);
                }
            }
        }

        private Image<Rgba, Byte> GetCVImageFromMFFrame( MFFrame _pFrame )
        {
            M_AV_PROPS avProps;
            int lAudioSamples;
            _pFrame.MFAVPropsGet(out avProps, out lAudioSamples);

            int nFrameWidth = avProps.vidProps.nWidth;
            int nFrameHeight = Math.Abs(avProps.vidProps.nHeight);
            int nFrameRowBytes = avProps.vidProps.nRowBytes;

            int pcbSize;
            long pbVideo;
            _pFrame.MFVideoGetBytes(out pcbSize, out pbVideo);

            IntPtr pnVideoData = new IntPtr(pbVideo);

            // Pixel format for OpenCV image and MFormats frame must be the same
            Image<Rgba, Byte> cvFrame = new Image<Rgba, Byte>(nFrameWidth, nFrameHeight, nFrameRowBytes, pnVideoData);

            return cvFrame;
        }


        private void FillMLiveDevices()
        {
            comboBoxVideo.Items.Clear();
            //Fill video devices
            int vCount;
            try
            {
                m_objLive = new MFLiveClass();

                m_objLive.DeviceGetCount(eMFDeviceType.eMFDT_Video, out vCount);

                if (vCount > 0)
                {
                    for (int i = 0; i < vCount; i++)
                    {
                        string strName;
                        int _pbBusy;
                        m_objLive.DeviceGetByIndex(eMFDeviceType.eMFDT_Video, i, out strName, out _pbBusy);
                        comboBoxVideo.Items.Add(strName);

                    }
                    comboBoxVideo.SelectedIndex = 0;
                    
                    buttonInit.Enabled = true;
                    buttonClose.Enabled = true;
                }

            }
            catch { }
        }

        private void comboBoxV_SelectedIndexChanged(object sender, EventArgs e)
        {
            try
            {
                comboBoxFormat.Items.Clear();
                comboBoxVideoLine.Items.Clear();

                M_VID_PROPS vidProps;

                m_objLive.DeviceSet(eMFDeviceType.eMFDT_Video, comboBoxVideo.SelectedIndex, "");

                buttonInit.Enabled = true;
                buttonClose.Enabled = true;

                int fCount, lCount;
                string name, _pOptions, _pHelp;

                m_objLive.FormatVideoGetCount(eMFormatType.eMFT_Input, out fCount);
                if (fCount > 0)
                {
                    for (int i = 0; i < fCount; i++)
                    {
                        m_objLive.FormatVideoGetByIndex(eMFormatType.eMFT_Input, i, out vidProps, out name);
                        comboBoxFormat.Items.Add(name);
                    }

                    // Get current format
                    int nCurrent = 0;
                    try
                    {
                        string strName;
                        M_VID_PROPS _vidProps;
                        m_objLive.FormatVideoGet(eMFormatType.eMFT_Input, out _vidProps, out nCurrent, out strName);
                    }
                    catch (System.Exception)
                    {
                    }

                    // Select current
                    comboBoxFormat.SelectedIndex = nCurrent > 0 ? nCurrent : 0;
                }

                string strCurrent;
                int nSel = -1;
                m_objLive.PropsGet("line-in", out strCurrent);

                m_objLive.PropsOptionGetCount("line-in", out lCount);
                if (lCount > 0)
                {
                    for (int i = 0; i < lCount; i++)
                    {
                        m_objLive.PropsOptionGetByIndex("line-in", i, out _pOptions, out _pHelp);
                        comboBoxVideoLine.Items.Add(_pHelp);

                        if (strCurrent == _pOptions)
                            nSel = i;
                    }

                    if (nSel >= 0)
                        comboBoxVideoLine.SelectedIndex = nSel;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        private void comboBoxVL_SelectedIndexChanged(object sender, EventArgs e)
        {
            m_objLive.PropsOptionSetByIndex("line-in", comboBoxVideoLine.SelectedIndex);
        }

        private void comboBoxVF_SelectedIndexChanged(object sender, EventArgs e)
        {
            M_VID_PROPS vidProps;
            string name;
            m_objLive.FormatVideoGetByIndex(eMFormatType.eMFT_Input, comboBoxFormat.SelectedIndex, out vidProps, out name);

            m_objLive.FormatVideoSet(eMFormatType.eMFT_Input, vidProps);
        }

        private void btnInit_Click(object sender, EventArgs e)
        {
            // Set device
            m_objLive.DeviceSet(eMFDeviceType.eMFDT_Video, comboBoxVideo.SelectedIndex, "");

            // Set format
            M_VID_PROPS vidProps;
            string name;
            m_objLive.FormatVideoGetByIndex(eMFormatType.eMFT_Input, comboBoxFormat.SelectedIndex, out vidProps, out name);
            m_objLive.FormatVideoSet(eMFormatType.eMFT_Input, vidProps);

            // Start worker
            m_bWork = true;
        }

        private void button2_Click(object sender, EventArgs e)
        {
            m_bWork = false;
            m_objLive.DeviceClose();
        }

        private void panelPreview_SizeChanged(object sender, EventArgs e)
        {
			if (m_objPreview != null )
				m_objPreview.PreviewWindowSet("", panelPreview.Handle.ToInt32());
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            cancelSource.Cancel();
            m_threadWorker.Join(1000);
			
            m_objLive.DeviceClose();
        }
    }
}
